# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from unittest.mock import AsyncMock, Mock, call, patch

import pytest
from llama_index.core.tools.types import ToolOutput
from pydantic import BaseModel
from toolbox_core.protocol import ParameterSchema as CoreParameterSchema
from toolbox_core.sync_tool import ToolboxSyncTool as ToolboxCoreSyncTool
from toolbox_core.tool import ToolboxTool as CoreAsyncTool
from toolbox_core.utils import params_to_pydantic_model

from toolbox_llamaindex.async_tools import AsyncToolboxTool
from toolbox_llamaindex.tools import ToolboxTool


def assert_pydantic_models_equivalent(
    model_cls1: type[BaseModel], model_cls2: type[BaseModel], expected_model_name: str
):
    assert issubclass(model_cls1, BaseModel), "model_cls1 is not a Pydantic BaseModel"
    assert issubclass(model_cls2, BaseModel), "model_cls2 is not a Pydantic BaseModel"
    assert (
        model_cls1.__name__ == expected_model_name
    ), f"model_cls1 name mismatch: expected {expected_model_name}, got {model_cls1.__name__}"
    assert (
        model_cls2.__name__ == expected_model_name
    ), f"model_cls2 name mismatch: expected {expected_model_name}, got {model_cls2.__name__}"

    fields1 = model_cls1.model_fields
    fields2 = model_cls2.model_fields

    assert (
        fields1.keys() == fields2.keys()
    ), f"Field names mismatch: {fields1.keys()} != {fields2.keys()}"

    for field_name in fields1.keys():
        field_info1 = fields1[field_name]
        field_info2 = fields2[field_name]

        assert (
            field_info1.annotation == field_info2.annotation
        ), f"Field '{field_name}': Annotation mismatch ({field_info1.annotation} != {field_info2.annotation})"
        assert (
            field_info1.description == field_info2.description
        ), f"Field '{field_name}': Description mismatch ('{field_info1.description}' != '{field_info2.description}')"
        is_required1 = (
            field_info1.is_required()
            if hasattr(field_info1, "is_required")
            else not field_info1.is_nullable()
        )
        is_required2 = (
            field_info2.is_required()
            if hasattr(field_info2, "is_required")
            else not field_info2.is_nullable()
        )
        assert (
            is_required1 == is_required2
        ), f"Field '{field_name}': Required status mismatch ({is_required1} != {is_required2})"


class TestToolboxTool:
    @pytest.fixture
    def tool_schema_dict(self):
        return {
            "description": "Test Tool Description",
            "parameters": [
                {"name": "param1", "type": "string", "description": "Param 1"},
                {"name": "param2", "type": "integer", "description": "Param 2"},
            ],
        }

    @pytest.fixture
    def auth_tool_schema_dict(self):
        return {
            "description": "Test Auth Tool Description",
            "authRequired": ["test-auth-source"],
            "parameters": [
                {
                    "name": "param1",
                    "type": "string",
                    "description": "Param 1",
                    "authSources": ["test-auth-source"],
                },
                {"name": "param2", "type": "integer", "description": "Param 2"},
            ],
        }

    @pytest.fixture
    def mock_core_tool(self, tool_schema_dict):
        sync_mock = Mock(spec=ToolboxCoreSyncTool)

        sync_mock.__name__ = "test_tool_name_for_llamaindex"
        sync_mock.__doc__ = tool_schema_dict["description"]
        sync_mock._name = "TestToolPydanticModel"
        sync_mock._params = [
            CoreParameterSchema(**p) for p in tool_schema_dict["parameters"]
        ]

        mock_async_tool_attr = AsyncMock(spec=CoreAsyncTool)
        mock_async_tool_attr.return_value = "dummy_internal_async_tool_result"
        sync_mock._ToolboxSyncTool__async_tool = mock_async_tool_attr
        sync_mock._ToolboxSyncTool__loop = Mock(spec=asyncio.AbstractEventLoop)
        sync_mock._ToolboxSyncTool__thread = Mock()

        new_mock_instance_for_methods = Mock(spec=ToolboxCoreSyncTool)
        new_mock_instance_for_methods.__name__ = sync_mock.__name__
        new_mock_instance_for_methods.__doc__ = sync_mock.__doc__
        new_mock_instance_for_methods._name = sync_mock._name
        new_mock_instance_for_methods._params = sync_mock._params
        new_mock_instance_for_methods._ToolboxSyncTool__async_tool = AsyncMock(
            spec=CoreAsyncTool
        )
        new_mock_instance_for_methods._ToolboxSyncTool__loop = Mock(
            spec=asyncio.AbstractEventLoop
        )
        new_mock_instance_for_methods._ToolboxSyncTool__thread = Mock()

        sync_mock.add_auth_token_getters = Mock(
            return_value=new_mock_instance_for_methods
        )
        sync_mock.bind_params = Mock(return_value=new_mock_instance_for_methods)

        return sync_mock

    @pytest.fixture
    def mock_core_sync_auth_tool(self, auth_tool_schema_dict):
        sync_mock = Mock(spec=ToolboxCoreSyncTool)
        sync_mock.__name__ = "test_auth_tool_lc_name"
        sync_mock.__doc__ = auth_tool_schema_dict["description"]
        sync_mock._name = "TestAuthToolPydanticModel"
        sync_mock._params = [
            CoreParameterSchema(**p) for p in auth_tool_schema_dict["parameters"]
        ]

        mock_async_tool_attr = AsyncMock(spec=CoreAsyncTool)
        mock_async_tool_attr.return_value = "dummy_internal_async_auth_tool_result"
        sync_mock._ToolboxSyncTool__async_tool = mock_async_tool_attr
        sync_mock._ToolboxSyncTool__loop = Mock(spec=asyncio.AbstractEventLoop)
        sync_mock._ToolboxSyncTool__thread = Mock()

        new_mock_instance_for_methods = Mock(spec=ToolboxCoreSyncTool)
        new_mock_instance_for_methods.__name__ = sync_mock.__name__
        new_mock_instance_for_methods.__doc__ = sync_mock.__doc__
        new_mock_instance_for_methods._name = sync_mock._name
        new_mock_instance_for_methods._params = sync_mock._params
        new_mock_instance_for_methods._ToolboxSyncTool__async_tool = AsyncMock(
            spec=CoreAsyncTool
        )
        new_mock_instance_for_methods._ToolboxSyncTool__loop = Mock(
            spec=asyncio.AbstractEventLoop
        )
        new_mock_instance_for_methods._ToolboxSyncTool__thread = Mock()

        sync_mock.add_auth_token_getters = Mock(
            return_value=new_mock_instance_for_methods
        )
        sync_mock.bind_params = Mock(return_value=new_mock_instance_for_methods)
        return sync_mock

    @pytest.fixture
    def toolbox_tool(self, mock_core_tool):
        return ToolboxTool(core_tool=mock_core_tool)

    @pytest.fixture
    def auth_toolbox_tool(self, mock_core_sync_auth_tool):
        return ToolboxTool(core_tool=mock_core_sync_auth_tool)

    def test_toolbox_tool_init(self, mock_core_tool):
        tool = ToolboxTool(core_tool=mock_core_tool)

        assert tool.metadata.name == mock_core_tool.__name__
        assert tool.metadata.description == mock_core_tool.__doc__
        assert tool._ToolboxTool__core_tool == mock_core_tool

        expected_args_schema = params_to_pydantic_model(
            mock_core_tool._name, mock_core_tool._params
        )
        assert_pydantic_models_equivalent(
            tool.metadata.fn_schema, expected_args_schema, mock_core_tool._name
        )

    @pytest.mark.parametrize(
        "params",
        [
            ({"param1": "bound-value"}),
            ({"param1": lambda: "bound-value"}),
            ({"param1": "bound-value", "param2": 123}),
        ],
    )
    def test_toolbox_tool_bind_params(
        self,
        params,
        toolbox_tool,
        mock_core_tool,
    ):
        returned_core_tool_mock = mock_core_tool.bind_params.return_value
        new_llamaindex_tool = toolbox_tool.bind_params(params)

        mock_core_tool.bind_params.assert_called_once_with(params)
        assert isinstance(new_llamaindex_tool, ToolboxTool)
        assert new_llamaindex_tool._ToolboxTool__core_tool == returned_core_tool_mock

    def test_toolbox_tool_bind_param(self, toolbox_tool, mock_core_tool):
        returned_core_tool_mock = mock_core_tool.bind_params.return_value
        new_llamaindex_tool = toolbox_tool.bind_param("param1", "bound-value")

        mock_core_tool.bind_params.assert_called_once_with({"param1": "bound-value"})
        assert isinstance(new_llamaindex_tool, ToolboxTool)
        assert new_llamaindex_tool._ToolboxTool__core_tool == returned_core_tool_mock

    @pytest.mark.parametrize(
        "auth_token_getters",
        [
            ({"test-auth-source": lambda: "test-token"}),
            (
                {
                    "test-auth-source": lambda: "test-token",
                    "another-auth-source": lambda: "another-token",
                }
            ),
        ],
    )
    def test_toolbox_tool_add_auth_token_getters(
        self,
        auth_token_getters,
        auth_toolbox_tool,
        mock_core_sync_auth_tool,
    ):
        returned_core_tool_mock = (
            mock_core_sync_auth_tool.add_auth_token_getters.return_value
        )
        new_llamaindex_tool = auth_toolbox_tool.add_auth_token_getters(
            auth_token_getters
        )

        mock_core_sync_auth_tool.add_auth_token_getters.assert_called_once_with(
            auth_token_getters
        )
        assert isinstance(new_llamaindex_tool, ToolboxTool)
        assert new_llamaindex_tool._ToolboxTool__core_tool == returned_core_tool_mock

    def test_toolbox_tool_add_auth_token_getter(
        self, auth_toolbox_tool, mock_core_sync_auth_tool
    ):
        get_id_token = lambda: "test-token"
        returned_core_tool_mock = (
            mock_core_sync_auth_tool.add_auth_token_getters.return_value
        )

        new_llamaindex_tool = auth_toolbox_tool.add_auth_token_getter(
            "test-auth-source", get_id_token
        )

        mock_core_sync_auth_tool.add_auth_token_getters.assert_called_once_with(
            {"test-auth-source": get_id_token}
        )
        assert isinstance(new_llamaindex_tool, ToolboxTool)
        assert new_llamaindex_tool._ToolboxTool__core_tool == returned_core_tool_mock

    def test_toolbox_tool_run(self, toolbox_tool, mock_core_tool):
        kwargs_to_run = {"param1": "run_value1", "param2": 100}
        expected_result = ToolOutput(
            content="sync_run_output",
            tool_name="test_tool_name_for_llamaindex",
            raw_input=kwargs_to_run,
            raw_output="sync_run_output",
            is_error=False,
        )
        mock_core_tool.return_value = "sync_run_output"

        result = toolbox_tool.call(**kwargs_to_run)

        assert result == expected_result
        assert mock_core_tool.call_count == 1
        assert mock_core_tool.call_args == call(**kwargs_to_run)

    @pytest.mark.asyncio
    @patch("toolbox_llamaindex.tools.to_thread", new_callable=AsyncMock)
    async def test_toolbox_tool_arun(
        self, mock_to_thread_in_tools, toolbox_tool, mock_core_tool
    ):
        kwargs_to_run = {"param1": "arun_value1", "param2": 200}
        expected_result = ToolOutput(
            content="async_run_output",
            tool_name="test_tool_name_for_llamaindex",
            raw_input=kwargs_to_run,
            raw_output="async_run_output",
            is_error=False,
        )

        mock_core_tool.return_value = "async_run_output"

        async def to_thread_side_effect(func, *args, **kwargs_for_func):
            return func(**kwargs_for_func)

        mock_to_thread_in_tools.side_effect = to_thread_side_effect

        result = await toolbox_tool.acall(**kwargs_to_run)

        assert result == expected_result
        mock_to_thread_in_tools.assert_awaited_once_with(
            mock_core_tool, **kwargs_to_run
        )

        assert mock_core_tool.call_count == 1
        assert mock_core_tool.call_args == call(**kwargs_to_run)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import types  # For MappingProxyType
from unittest.mock import AsyncMock, Mock, patch

import pytest
import pytest_asyncio
from llama_index.core.tools.types import ToolOutput
from pydantic import ValidationError
from toolbox_core.protocol import ParameterSchema as CoreParameterSchema
from toolbox_core.tool import ToolboxTool as ToolboxCoreTool

from toolbox_llamaindex.async_tools import AsyncToolboxTool


@pytest.mark.asyncio
class TestAsyncToolboxTool:
    @pytest.fixture
    def tool_schema_dict(self):
        return {
            "description": "Test Tool Description",
            "parameters": [
                {"name": "param1", "type": "string", "description": "Param 1"},
                {"name": "param2", "type": "integer", "description": "Param 2"},
            ],
        }

    @pytest.fixture
    def auth_tool_schema_dict(self):
        return {
            "description": "Test Tool Description",
            "authRequired": ["test-auth-source"],
            "parameters": [
                {
                    "name": "param1",
                    "type": "string",
                    "description": "Param 1",
                    "authSources": ["test-auth-source"],
                },
                {"name": "param2", "type": "integer", "description": "Param 2"},
            ],
        }

    def _create_core_tool_from_dict(
        self, session, name, schema_dict, url, initial_auth_getters=None
    ):
        core_params_schemas = [
            CoreParameterSchema(**p) for p in schema_dict["parameters"]
        ]

        tool_constructor_params = []
        required_authn_for_core = {}
        for p_schema in core_params_schemas:
            if p_schema.authSources:
                required_authn_for_core[p_schema.name] = p_schema.authSources
            else:
                tool_constructor_params.append(p_schema)

        return ToolboxCoreTool(
            session=session,
            base_url=url,
            name=name,
            description=schema_dict["description"],
            params=tool_constructor_params,
            required_authn_params=types.MappingProxyType(required_authn_for_core),
            required_authz_tokens=schema_dict.get("authRequired", []),
            auth_service_token_getters=types.MappingProxyType(
                initial_auth_getters or {}
            ),
            bound_params=types.MappingProxyType({}),
            client_headers=types.MappingProxyType({}),
        )

    @pytest_asyncio.fixture
    @patch("aiohttp.ClientSession")
    async def toolbox_tool(self, MockClientSession, tool_schema_dict):
        mock_session = MockClientSession.return_value
        mock_response = mock_session.post.return_value.__aenter__.return_value
        mock_response.raise_for_status = Mock()
        mock_response.json = AsyncMock(return_value={"result": "test-result"})
        mock_response.status = 200  # *** Fix: Set status for the mock response ***

        core_tool_instance = self._create_core_tool_from_dict(
            session=mock_session,
            name="test_tool",
            schema_dict=tool_schema_dict,
            url="http://test_url",
        )
        tool = AsyncToolboxTool(core_tool=core_tool_instance)
        return tool

    @pytest_asyncio.fixture
    @patch("aiohttp.ClientSession")
    async def auth_toolbox_tool(self, MockClientSession, auth_tool_schema_dict):
        mock_session = MockClientSession.return_value
        mock_response = mock_session.post.return_value.__aenter__.return_value
        mock_response.raise_for_status = Mock()
        mock_response.json = AsyncMock(return_value={"result": "test-result"})
        mock_response.status = 200  # *** Fix: Set status for the mock response ***

        core_tool_instance = self._create_core_tool_from_dict(
            session=mock_session,
            name="test_tool",
            schema_dict=auth_tool_schema_dict,
            url="https://test-url",
        )
        tool = AsyncToolboxTool(core_tool=core_tool_instance)
        return tool

    @patch("aiohttp.ClientSession")
    async def test_toolbox_tool_init(self, MockClientSession, tool_schema_dict):
        mock_session = MockClientSession.return_value
        mock_response = mock_session.post.return_value.__aenter__.return_value
        mock_response.status = 200
        core_tool_instance = self._create_core_tool_from_dict(
            session=mock_session,
            name="test_tool",
            schema_dict=tool_schema_dict,
            url="https://test-url",
        )
        tool = AsyncToolboxTool(core_tool=core_tool_instance)
        assert tool.metadata.name == "test_tool"
        assert tool.metadata.description == core_tool_instance.__doc__

    @pytest.mark.parametrize(
        "params_to_bind",
        [
            ({"param1": "bound-value"}),
            ({"param1": lambda: "bound-value"}),
            ({"param1": "bound-value", "param2": 123}),
        ],
    )
    async def test_toolbox_tool_bind_params(self, toolbox_tool, params_to_bind):
        original_core_tool = toolbox_tool._AsyncToolboxTool__core_tool
        with patch.object(
            original_core_tool, "bind_params", wraps=original_core_tool.bind_params
        ) as mock_core_bind_params:
            new_llamaindex_tool = toolbox_tool.bind_params(params_to_bind)
            mock_core_bind_params.assert_called_once_with(params_to_bind)
            assert isinstance(
                new_llamaindex_tool._AsyncToolboxTool__core_tool, ToolboxCoreTool
            )
            new_core_tool_signature_params = (
                new_llamaindex_tool._AsyncToolboxTool__core_tool.__signature__.parameters
            )
            for bound_param_name in params_to_bind.keys():
                assert bound_param_name not in new_core_tool_signature_params

    async def test_toolbox_tool_bind_params_invalid(self, toolbox_tool):
        with pytest.raises(
            ValueError, match="unable to bind parameters: no parameter named param3"
        ):
            toolbox_tool.bind_params({"param3": "bound-value"})

    async def test_toolbox_tool_bind_params_duplicate(self, toolbox_tool):
        tool = toolbox_tool.bind_params({"param1": "bound-value"})
        with pytest.raises(
            ValueError,
            match="cannot re-bind parameter: parameter 'param1' is already bound",
        ):
            tool.bind_params({"param1": "bound-value"})

    async def test_toolbox_tool_bind_params_invalid_params(self, auth_toolbox_tool):
        auth_core_tool = auth_toolbox_tool._AsyncToolboxTool__core_tool
        # Verify that 'param1' is not in the list of bindable parameters for the core tool
        # because it requires authentication.
        assert "param1" not in [p.name for p in auth_core_tool._ToolboxTool__params]
        with pytest.raises(
            ValueError, match="unable to bind parameters: no parameter named param1"
        ):
            auth_toolbox_tool.bind_params({"param1": "bound-value"})

    async def test_toolbox_tool_add_valid_auth_token_getter(self, auth_toolbox_tool):
        get_token_lambda = lambda: "test-token-value"
        original_core_tool = auth_toolbox_tool._AsyncToolboxTool__core_tool
        with patch.object(
            original_core_tool,
            "add_auth_token_getters",
            wraps=original_core_tool.add_auth_token_getters,
        ) as mock_core_add_getters:
            tool = auth_toolbox_tool.add_auth_token_getters(
                {"test-auth-source": get_token_lambda}
            )
            mock_core_add_getters.assert_called_once_with(
                {"test-auth-source": get_token_lambda}
            )
            core_tool_after_add = tool._AsyncToolboxTool__core_tool
            assert (
                "test-auth-source"
                in core_tool_after_add._ToolboxTool__auth_service_token_getters
            )
            assert (
                core_tool_after_add._ToolboxTool__auth_service_token_getters[
                    "test-auth-source"
                ]
                is get_token_lambda
            )
            assert not core_tool_after_add._ToolboxTool__required_authn_params.get(
                "param1"
            )
            assert (
                "test-auth-source"
                not in core_tool_after_add._ToolboxTool__required_authz_tokens
            )

    async def test_toolbox_tool_add_unused_auth_token_getter_raises_error(
        self, auth_toolbox_tool
    ):
        unused_lambda = lambda: "another-token"
        with pytest.raises(ValueError) as excinfo:
            auth_toolbox_tool.add_auth_token_getters(
                {"another-auth-source": unused_lambda}
            )
        assert (
            "Authentication source(s) `another-auth-source` unused by tool `test_tool`"
            in str(excinfo.value)
        )

        valid_lambda = lambda: "test-token"
        with pytest.raises(ValueError) as excinfo_mixed:
            auth_toolbox_tool.add_auth_token_getters(
                {
                    "test-auth-source": valid_lambda,
                    "another-auth-source": unused_lambda,
                }
            )
        assert (
            "Authentication source(s) `another-auth-source` unused by tool `test_tool`"
            in str(excinfo_mixed.value)
        )

    async def test_toolbox_tool_add_auth_token_getters_duplicate(
        self, auth_toolbox_tool
    ):
        tool = auth_toolbox_tool.add_auth_token_getters(
            {"test-auth-source": lambda: "test-token"}
        )
        with pytest.raises(
            ValueError,
            match="Authentication source\\(s\\) `test-auth-source` already registered in tool `test_tool`\\.",
        ):
            tool.add_auth_token_getters({"test-auth-source": lambda: "test-token"})

    async def test_toolbox_tool_call_requires_auth_strict(self, auth_toolbox_tool):
        with pytest.raises(
            PermissionError,
            match="One or more of the following authn services are required to invoke this tool: test-auth-source",
        ):
            await auth_toolbox_tool.acall(param2=123)

    async def test_toolbox_tool_call(self, toolbox_tool):
        result = await toolbox_tool.acall(param1="test-value", param2=123)
        assert result == ToolOutput(
            content="test-result",
            tool_name="test_tool",
            raw_input={"param1": "test-value", "param2": 123},
            raw_output="test-result",
            is_error=False,
        )

        core_tool = toolbox_tool._AsyncToolboxTool__core_tool
        core_tool._ToolboxTool__session.post.assert_called_once_with(
            "http://test_url/api/tool/test_tool/invoke",
            json={"param1": "test-value", "param2": 123},
            headers={},
        )

    @pytest.mark.parametrize(
        "bound_param_map, expected_value",
        [
            ({"param1": "bound-value"}, "bound-value"),
            ({"param1": lambda: "dynamic-value"}, "dynamic-value"),
        ],
    )
    async def test_toolbox_tool_call_with_bound_params(
        self, toolbox_tool, bound_param_map, expected_value
    ):
        tool = toolbox_tool.bind_params(bound_param_map)
        result = await tool.acall(param2=123)
        assert result == ToolOutput(
            content="test-result",
            tool_name="test_tool",
            raw_input={"param2": 123},
            raw_output="test-result",
            is_error=False,
        )
        core_tool = tool._AsyncToolboxTool__core_tool
        core_tool._ToolboxTool__session.post.assert_called_once_with(
            "http://test_url/api/tool/test_tool/invoke",
            json={"param1": expected_value, "param2": 123},
            headers={},
        )

    async def test_toolbox_tool_call_with_auth_tokens(self, auth_toolbox_tool):
        tool = auth_toolbox_tool.add_auth_token_getters(
            {"test-auth-source": lambda: "test-token"}
        )
        result = await tool.acall(param2=123)
        assert result == ToolOutput(
            content="test-result",
            tool_name="test_tool",
            raw_input={"param2": 123},
            raw_output="test-result",
            is_error=False,
        )

        core_tool = tool._AsyncToolboxTool__core_tool
        core_tool._ToolboxTool__session.post.assert_called_once_with(
            "https://test-url/api/tool/test_tool/invoke",
            json={"param2": 123},
            headers={"test-auth-source_token": "test-token"},
        )

    async def test_toolbox_tool_call_with_auth_tokens_insecure(
        self, auth_toolbox_tool, auth_tool_schema_dict
    ):
        core_tool_of_auth_tool = auth_toolbox_tool._AsyncToolboxTool__core_tool
        mock_session = core_tool_of_auth_tool._ToolboxTool__session

        with pytest.warns(
            UserWarning,
            match="Sending ID token over HTTP. User data may be exposed. Use HTTPS for secure communication.",
        ):
            insecure_core_tool = self._create_core_tool_from_dict(
                session=mock_session,
                name="test_tool",
                schema_dict=auth_tool_schema_dict,
                url="http://test-url",
            )

        insecure_auth_langchain_tool = AsyncToolboxTool(core_tool=insecure_core_tool)

        tool_with_getter = insecure_auth_langchain_tool.add_auth_token_getters(
            {"test-auth-source": lambda: "test-token"}
        )
        result = await tool_with_getter.acall(param2=123)
        assert result == ToolOutput(
            content="test-result",
            tool_name="test_tool",
            raw_input={"param2": 123},
            raw_output="test-result",
            is_error=False,
        )

        modified_core_tool_in_new_tool = tool_with_getter._AsyncToolboxTool__core_tool
        assert (
            modified_core_tool_in_new_tool._ToolboxTool__base_url == "http://test-url"
        )
        assert (
            modified_core_tool_in_new_tool._ToolboxTool__url
            == "http://test-url/api/tool/test_tool/invoke"
        )

        modified_core_tool_in_new_tool._ToolboxTool__session.post.assert_called_once_with(
            "http://test-url/api/tool/test_tool/invoke",
            json={"param2": 123},
            headers={"test-auth-source_token": "test-token"},
        )

    async def test_toolbox_tool_call_with_empty_input(self, toolbox_tool):
        with pytest.raises(TypeError) as e:
            await toolbox_tool.acall()
        assert "missing a required argument: 'param1'" in str(e.value)

    async def test_toolbox_tool_call_with_invalid_input(self, toolbox_tool):
        with pytest.raises(ValidationError) as e:
            await toolbox_tool.acall(param1=123, param2="invalid")
        assert "2 validation errors for test_tool" in str(e.value)
        assert "param1\n  Input should be a valid string" in str(e.value)
        assert "param2\n  Input should be a valid integer" in str(e.value)

    async def test_toolbox_tool_run_not_implemented(self, toolbox_tool):
        with pytest.raises(NotImplementedError):
            toolbox_tool.call()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asyncio import to_thread
from typing import Any, Awaitable, Callable, Mapping, Optional, Union
from warnings import warn

from toolbox_core.sync_client import ToolboxSyncClient as ToolboxCoreSyncClient
from toolbox_core.sync_tool import ToolboxSyncTool

from .tools import ToolboxTool


class ToolboxClient:

    def __init__(
        self,
        url: str,
        client_headers: Optional[
            Mapping[str, Union[Callable[[], str], Callable[[], Awaitable[str]], str]]
        ] = None,
    ) -> None:
        """
        Initializes the ToolboxClient for the Toolbox service at the given URL.

        Args:
            url: The base URL of the Toolbox service.
        """
        self.__core_client = ToolboxCoreSyncClient(
            url=url, client_headers=client_headers
        )

    async def aload_tool(
        self,
        tool_name: str,
        auth_token_getters: dict[str, Callable[[], str]] = {},
        auth_tokens: Optional[dict[str, Callable[[], str]]] = None,
        auth_headers: Optional[dict[str, Callable[[], str]]] = None,
        bound_params: dict[str, Union[Any, Callable[[], Any]]] = {},
    ) -> ToolboxTool:
        """
        Loads the tool with the given tool name from the Toolbox service.

        Args:
            tool_name: The name of the tool to load.
            auth_token_getters: An optional mapping of authentication source
                names to functions that retrieve ID tokens.
            auth_tokens: Deprecated. Use `auth_token_getters` instead.
            auth_headers: Deprecated. Use `auth_token_getters` instead.
            bound_params: An optional mapping of parameter names to their
                bound values.

        Returns:
            A tool loaded from the Toolbox.
        """
        if auth_tokens:
            if auth_token_getters:
                warn(
                    "Both `auth_token_getters` and `auth_tokens` are provided. `auth_tokens` is deprecated, and `auth_token_getters` will be used.",
                    DeprecationWarning,
                )
            else:
                warn(
                    "Argument `auth_tokens` is deprecated. Use `auth_token_getters` instead.",
                    DeprecationWarning,
                )
                auth_token_getters = auth_tokens

        if auth_headers:
            if auth_token_getters:
                warn(
                    "Both `auth_token_getters` and `auth_headers` are provided. `auth_headers` is deprecated, and `auth_token_getters` will be used.",
                    DeprecationWarning,
                )
            else:
                warn(
                    "Argument `auth_headers` is deprecated. Use `auth_token_getters` instead.",
                    DeprecationWarning,
                )
                auth_token_getters = auth_headers

        core_tool = await to_thread(
            self.__core_client.load_tool,
            name=tool_name,
            auth_token_getters=auth_token_getters,
            bound_params=bound_params,
        )
        return ToolboxTool(core_tool=core_tool)

    async def aload_toolset(
        self,
        toolset_name: Optional[str] = None,
        auth_token_getters: dict[str, Callable[[], str]] = {},
        auth_tokens: Optional[dict[str, Callable[[], str]]] = None,
        auth_headers: Optional[dict[str, Callable[[], str]]] = None,
        bound_params: dict[str, Union[Any, Callable[[], Any]]] = {},
        strict: bool = False,
    ) -> list[ToolboxTool]:
        """
        Loads tools from the Toolbox service, optionally filtered by toolset
        name.

        Args:
            toolset_name: The name of the toolset to load. If not provided,
                all tools are loaded.
            auth_token_getters: An optional mapping of authentication source
                names to functions that retrieve ID tokens.
            auth_tokens: Deprecated. Use `auth_token_getters` instead.
            auth_headers: Deprecated. Use `auth_token_getters` instead.
            bound_params: An optional mapping of parameter names to their
                bound values.
            strict: If True, raises an error if *any* loaded tool instance fails
                to utilize at least one provided parameter or auth token (if any
                provided). If False (default), raises an error only if a
                user-provided parameter or auth token cannot be applied to *any*
                loaded tool across the set.

        Returns:
            A list of all tools loaded from the Toolbox.
        """
        if auth_tokens:
            if auth_token_getters:
                warn(
                    "Both `auth_token_getters` and `auth_tokens` are provided. `auth_tokens` is deprecated, and `auth_token_getters` will be used.",
                    DeprecationWarning,
                )
            else:
                warn(
                    "Argument `auth_tokens` is deprecated. Use `auth_token_getters` instead.",
                    DeprecationWarning,
                )
                auth_token_getters = auth_tokens

        if auth_headers:
            if auth_token_getters:
                warn(
                    "Both `auth_token_getters` and `auth_headers` are provided. `auth_headers` is deprecated, and `auth_token_getters` will be used.",
                    DeprecationWarning,
                )
            else:
                warn(
                    "Argument `auth_headers` is deprecated. Use `auth_token_getters` instead.",
                    DeprecationWarning,
                )
                auth_token_getters = auth_headers

        core_tools = await to_thread(
            self.__core_client.load_toolset,
            name=toolset_name,
            auth_token_getters=auth_token_getters,
            bound_params=bound_params,
            strict=strict,
        )

        tools = []
        for core_tool in core_tools:
            tools.append(ToolboxTool(core_tool=core_tool))
        return tools

    def load_tool(
        self,
        tool_name: str,
        auth_token_getters: dict[str, Callable[[], str]] = {},
        auth_tokens: Optional[dict[str, Callable[[], str]]] = None,
        auth_headers: Optional[dict[str, Callable[[], str]]] = None,
        bound_params: dict[str, Union[Any, Callable[[], Any]]] = {},
    ) -> ToolboxTool:
        """
        Loads the tool with the given tool name from the Toolbox service.

        Args:
            tool_name: The name of the tool to load.
            auth_token_getters: An optional mapping of authentication source
                names to functions that retrieve ID tokens.
            auth_tokens: Deprecated. Use `auth_token_getters` instead.
            auth_headers: Deprecated. Use `auth_token_getters` instead.
            bound_params: An optional mapping of parameter names to their
                bound values.

        Returns:
            A tool loaded from the Toolbox.
        """
        if auth_tokens:
            if auth_token_getters:
                warn(
                    "Both `auth_token_getters` and `auth_tokens` are provided. `auth_tokens` is deprecated, and `auth_token_getters` will be used.",
                    DeprecationWarning,
                )
            else:
                warn(
                    "Argument `auth_tokens` is deprecated. Use `auth_token_getters` instead.",
                    DeprecationWarning,
                )
                auth_token_getters = auth_tokens

        if auth_headers:
            if auth_token_getters:
                warn(
                    "Both `auth_token_getters` and `auth_headers` are provided. `auth_headers` is deprecated, and `auth_token_getters` will be used.",
                    DeprecationWarning,
                )
            else:
                warn(
                    "Argument `auth_headers` is deprecated. Use `auth_token_getters` instead.",
                    DeprecationWarning,
                )
                auth_token_getters = auth_headers

        core_sync_tool = self.__core_client.load_tool(
            name=tool_name,
            auth_token_getters=auth_token_getters,
            bound_params=bound_params,
        )
        return ToolboxTool(core_tool=core_sync_tool)

    def load_toolset(
        self,
        toolset_name: Optional[str] = None,
        auth_token_getters: dict[str, Callable[[], str]] = {},
        auth_tokens: Optional[dict[str, Callable[[], str]]] = None,
        auth_headers: Optional[dict[str, Callable[[], str]]] = None,
        bound_params: dict[str, Union[Any, Callable[[], Any]]] = {},
        strict: bool = False,
    ) -> list[ToolboxTool]:
        """
        Loads tools from the Toolbox service, optionally filtered by toolset
        name.

        Args:
            toolset_name: The name of the toolset to load. If not provided,
                all tools are loaded.
            auth_token_getters: An optional mapping of authentication source
                names to functions that retrieve ID tokens.
            auth_tokens: Deprecated. Use `auth_token_getters` instead.
            auth_headers: Deprecated. Use `auth_token_getters` instead.
            bound_params: An optional mapping of parameter names to their
                bound values.
            strict: If True, raises an error if *any* loaded tool instance fails
                to utilize at least one provided parameter or auth token (if any
                provided). If False (default), raises an error only if a
                user-provided parameter or auth token cannot be applied to *any*
                loaded tool across the set.

        Returns:
            A list of all tools loaded from the Toolbox.
        """
        if auth_tokens:
            if auth_token_getters:
                warn(
                    "Both `auth_token_getters` and `auth_tokens` are provided. `auth_tokens` is deprecated, and `auth_token_getters` will be used.",
                    DeprecationWarning,
                )
            else:
                warn(
                    "Argument `auth_tokens` is deprecated. Use `auth_token_getters` instead.",
                    DeprecationWarning,
                )
                auth_token_getters = auth_tokens

        if auth_headers:
            if auth_token_getters:
                warn(
                    "Both `auth_token_getters` and `auth_headers` are provided. `auth_headers` is deprecated, and `auth_token_getters` will be used.",
                    DeprecationWarning,
                )
            else:
                warn(
                    "Argument `auth_headers` is deprecated. Use `auth_token_getters` instead.",
                    DeprecationWarning,
                )
                auth_token_getters = auth_headers

        core_sync_tools = self.__core_client.load_toolset(
            name=toolset_name,
            auth_token_getters=auth_token_getters,
            bound_params=bound_params,
            strict=strict,
        )

        tools = []
        for core_sync_tool in core_sync_tools:
            tools.append(ToolboxTool(core_tool=core_sync_tool))
        return tools

    def close(self):
        """Close the underlying synchronous client."""
        self.__core_client.close()

    async def __aenter__(self):
        """
        Enter the runtime context related to this client instance.

        Allows the client to be used as an asynchronous context manager
        (e.g., `async with ToolboxClient(...) as client:`).

        Returns:
            self: The client instance itself.
        """
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Exit the runtime context and close the internally managed session.

        Allows the client to be used as an asynchronous context manager
        (e.g., `async with ToolboxClient(...) as client:`).
        """
        self.close()

    def __enter__(self):
        """
        Enter the runtime context related to this client instance.

        Allows the client to be used as a context manager
        (e.g., `with ToolboxClient(...) as client:`).

        Returns:
            self: The client instance itself.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exit the runtime context and close the internally managed session.

        Allows the client to be used as a context manager
        (e.g., `with ToolboxClient(...) as client:`).
        """
        self.close()
